"""
    This module contains the entry points to the manifest-edit python
    application.

    These are two, and must be invoked in order:

    1. a "setup" method, to prepare a processing pipeline by loading python
    plugins from disk based on a .yaml configuration file
    2. an "update" method, to process a manifest through the pipeline

    It must be noticed that the manifest is expected to be represented with the
    appropriate libfmp4.mpd.Manifest type.

    For details, please refere to the individual methods' documentation.

    All other methods are private and not supposed to be invoked from other
    modules.
"""

from manifest_edit import libfmp4
from .pipeline import Pipeline
from .context import Context
from .libfmp4.utility import copyStdVectors, copyPythonToCppMembers

_comment_header = "Edited by Unified manifest-edit, using the following "\
                  "pipeline:\n"


def _addManifestEditComment(manifest, pipeline_config):
    # FIXME: add comment field to m3u8 manifest
    if type(manifest) == libfmp4.hls.Manifest:
        return
    # These operations need to be executed once only (at startup time). Only
    # after executing this method you are ready to receive manifests for
    # editing
    manifest.comment = _comment_header + pipeline_config


def setup(mp4process_context, pipeline_filename):
    """
    This method prepares a processing pipeline for a specific manifest
    format (in this version, only "mpd" is supported).

    This method must be called at least once before invoking "update_manifest".

    For efficiency purposes, if multiple manifests need to be processed, the
    calling module should invoke "setup" just once to prepare the pipeline and
    then invoke multiple times "update_manifest".

    However, this method is optimized to load a certain .yaml file from disk
    just the first time it is encountered and then cache and reuse the
    resulting pipeline in subsequent calls, making it possible to repeatedly
    call "setup" with different pipeline file names with low impact on
    performance.

    Args:
        - extension (str): The manifest extension ("mpd" supported only)

        - mp4process_context (libfmp4.Mp4ProcessContext): an mp4process
          context used to generate the .mpd file. Only used to access the
          logging functions and share them with the calling module.
          You can pass None to disable log sharing and default to print-based
          logging.

        - pipeline_filename (str): the name of the .yaml Pipeline Configuration
          file.

    Returns:
        None.

    Raises:
        Exception: for unsupported extension.
        FileNotFoundError: for pipeline configuration file not found.
        yaml.YAMLError: for pipeline configuration file yaml syntax error.
        ImportError: if a plugin mentioned in the pipeline configuration file
            cannot be found/imported.
        SchemaError: for invalid plugin configuration.
    """

    Context.mp4context = mp4process_context
    _setup(pipeline_filename)
    Context.log_debug("Setup with pipeline filename {} done.".format(
            pipeline_filename
        )
    )


# This version of setup is only intended for unit tests, where an mp4context
# is not available
def _setup(pipeline_filename, use_python_logs=False):
    Context.use_python_logs = use_python_logs

    Context.log_trace("Using libfmp4 = {}".format(libfmp4.version_string()))

    # build_pipeline will raise in case the pipeline cannot be built.
    # Also notice that this method is memoized: calling it several times
    #  with the same pipeline_filename will reuse pipelines from an internal
    #  cache instead of re-reading the yml file from disk.
    Context.pipeline = Pipeline.build_pipeline(pipeline_filename)


# this is the entry point for updating an MPEG DASH manifest
# def update_manifest(manifest: libfmp4.mpd.Manifest) -> libfmp4.mpd.Manifest:


def update_manifest(manifest):
    """
    This method processes the manifest through the active pipeline previously
    selected by the "setup" method.

    Args:
        - manifest (libfmp4.mpd.Manifest): The input manifest ("mpd" dash
          format only is supported)

    Returns:
        - the edited manifest.

    Raises:
        - NotImplementedError: if a "Plugin" class not respecting the required
          interface has been included in a pipeline.
        - Exception: in case of processing errors or unexpected manifest
          format.

    In details, this method performs:
        - dynamic identification of std:vector members (through the leading
          underscore convention used by libfmp4 python bindings and class
          reflection)
        - recursive copy of std:vector members to corresponding python list
          members
        - manifest process through pipeline
        - recursive copy-back of python list content to std:vector members
        - removal of the added python list from manifest

    """

    Context.log_trace("entering update manifest")
    # This call will dynamically add native python list members for every
    # std::vector in the manifest

    Context.log_trace("Copying STL containers")
    copyStdVectors(manifest)

    Context.log_trace("Pipeline processing...")
    # Actual pipeline processing
    Context.pipeline.run(input_manifest=manifest)

    # This call will copy python lists back into the original stl containers
    # and will remove dynamically added members
    Context.log_trace("Removing copied STL containers")
    copyPythonToCppMembers(manifest)

    # Finally, this will add a comment to the manifest with the specific
    #  yml description of the applied pipeline, mainly for troubleshooting
    #  or support purposes.
    Context.log_trace("Marking manifest as edited from manifest-edit")
    _addManifestEditComment(manifest, Context.pipeline.yaml_configuration)

    Context.log_trace("exiting update manifest")

    return manifest
