"""
manifest_edit.libfmp4.mpd

This package contains functionalities that are specific to .mpd (DASH)
manifests manipulation.

It contains data types that represent elements and attributes of a Media
Presentation as described by ISO/IEC FDIS 23009-1 4th edition.

When working with libfmp4.mpd, you abstract from the actual textual
representation format of the manifest (e.g. XML format) and deal directly with
nested native python lists to represent the manifest hierarchy.

At the moment, manifest-edit.libfmp4 does not expose to python any supported
serialization or deserialization methods, i.e. you cannot load and parse an
.mpd file to a libfmp4.mpd.Manifest data type or dump a libfmp4.mpd Manifest
object to a file. Its main purpose is to be used within the manifest-edit
application that has explicit entrypoints to receive and return
libfmp4.mpd.Manifest objects to/from an external applications. All methods
included in this library to render object to strings are only to be used for
debug purposes and are not suited to produce a legal XML representation of a
manifest or any of its parts.

It is worth remembering this package's naming convention:

 - free functions use lower snake case (e.g. my_function)
 - class names use upper CamelCase convention (e.g. MyClass)
 - class attributes use lower camelCase convention (e.g. myAttribute).

When using manifest-edit.libfmp4.mpd data types, some effort has been put to
make it usable with simple mnemonic rules to find types and members that you
may need. In particular, this library's classes and attributes names default to
the same names specified in ISO/IEC FDIS 23009-1 (save for the lower camelCase
naming convention). Some examples:

 - the "Period" element (ISO/IEC FDIS 23009-1 5.3.2) is represented by the
   "libfmp4.mpd.Period" data type. The "AdaptationSet" element
   (ISO/IEC FDIS 23009-1 5.3.3) is represented by the
   "libfmp4.mpd.AdaptationSet" data type. A notable exception to this rule
   is that the root "MPD" element is represented in the "libfmp4.mpd.Manifest"
   data type.

 - the "id" attribute of the "MPD" element is represented by the "id" attribute
   of the Manifest data type; the "start" attribute of a "Period" element is
   represented by the "start" attribute of the Period data type.

 - the "BaseURL" element of the "MPD" element is represented by the "baseURL"
   attribute of the "libfmp4.mpd.Manifest" data type: notice the naming
   convention with lower capital initial letter. Of course this has no
   influence on the string that is going to be included in the XML
   representation of the manifest, which will always reflect what specified in
   the standard.

Another general naming rule to keep in mind is that wherever the standard
specifies that an element can occur more than once (i.e. "0 ... N" is
specified), you are going to find these occurrences in a python list member,
with the same name of the element in the plural form. Some examples:

 - you will find all "Period" elements included in "MPD" in the member
   "periods" of the "libfmp4.mpd.Manifest" data type

 - you will find all "AdaptationSet" elements included in a "Period" in the
   member "adaptationSets" of the "libfmp4.mpd.Period" data type

A last but very important remark when using this library is about creation
of instances of this library's types. Even though constructors are directly
usable, it is required that each object is created by making use of the
``libfmp4.create`` factory class. This means you should always use

.. code-block:: python

  import manifest_edit.libfmp4.mpd as mpd
  my_period = create(mpd.Period)

``instead of``

.. code-block:: python

  import manifest_edit.libfmp4.mpd as mpd
  my_period = mpd.Period()

In future versions of the library the create factory method will probably be
made unnecessary but for the moment it is required.
"""
# Explicitly importing only those things that I want users to be exposed to.
from libfmp4.mpd import (  # noqa: F401
    Manifest,
    Period,
    AdaptationSet,
    Representation,
    ServiceDescription,
    SegmentTimeline,
    EventStream,
    Event,
    Descriptor,
    Latency,
    PlaybackRate,
)

from manifest_edit.libfmp4.utility import copyStdVectors as _copyStdVectors

# Importing as with underscode name to avoid autodoc to pick these imports up
from manifest_edit.libfmp4.utility import (
    deepcopy_align_and_add_python_lists as _deepcopy_align_and_add_python_lists
)

# This is what copy.copy() invokes. Automatically applies the copy/remove
# python members.
# Cannot do that for manifests because the needed C++ copy constructor is
# explicitly deleted in libfmp4
#setattr(Manifest, "__copy__", _deepcopy_align_and_add_python_lists)
setattr(Period, "__copy__", _deepcopy_align_and_add_python_lists)
setattr(AdaptationSet, "__copy__", _deepcopy_align_and_add_python_lists)
setattr(Representation, "__copy__", _deepcopy_align_and_add_python_lists)
setattr(ServiceDescription, "__copy__", _deepcopy_align_and_add_python_lists)
setattr(EventStream, "__copy__", _deepcopy_align_and_add_python_lists)
setattr(Event, "__copy__", _deepcopy_align_and_add_python_lists)

# These _copyStdVectors are invoked on the class object just to add
#  class attributes that are picked up by autodoc, for documentation
#  purposes. They have no other use and instance attributes are used only
#  anywhere else.
_copyStdVectors(Manifest)
_copyStdVectors(Period)
_copyStdVectors(AdaptationSet)
_copyStdVectors(Representation)
_copyStdVectors(ServiceDescription)
_copyStdVectors(EventStream)
_copyStdVectors(Event)
_copyStdVectors(Descriptor)
_copyStdVectors(Latency)
# _copyStdVectors(OptionalLatency)
_copyStdVectors(PlaybackRate)
