from manifest_edit.context import Context
import inspect
import sys

from .mpd import AdaptationSet, Representation # noqa: F401

# AdaptationSet and Representation has some attributes and elements in the
# "common" member. This list is here just to support a trick to allow users to
# do
#
#   adset.audioChannelsConfiguration = <...>
#
# instead of
#
#   adset.common.audioChannelsConfiguration = <...>
#
# It's not even an usability issue, more that for DASH, we established the
# convention that there is a one-to-one relationship to the manifest xml
# element/attribute name and the corresponding member name of the data
# structure provided by py-libfmp4.
# So, core parts of the code expects, i.e., that for a yaml selecting a period
# with
#
#   periods:
#     - id: "1"
#
# you can safely call
#
#   getattr(periods, "id")
#
# It seemed very easy initially to do so by just adding a property in
# fmp4_py.cpp that would just reference the similarly-named member in
# "common". We later discovered that this causes all sorts of memory issues,
# despite being seemingly supported.
# We now add the property on python, in the "copyStdVectors" method, and we
# remove it at the end in the "copyPythonToCppMember" method. This does not
# cause any memory issue because it's totally handled on python side.
#
# This trick will eventually be removed in favour of a proper
# "attribute-to-member-mapping" approach as used in HLS already.
#
# Any AdaptationSet or Representation member present in this list, will be
# copied from "common" to the adaptationset/representation member
#
# Obviously this must always be aligned with the members in RepresentationBase
COMMON_MEMBERS = [
    "width",
    "height",
    "sar",
    "frameRate",
    "audioSamplingRate",
    "mimeType",
    "codecs",
    "startWithSAP",
    "maxPlayoutRate",
    "codingDependency",
    "scanType",    
    "supplementalProperties",
    "essentialProperties",
    "audioChannelConfigurations",
]


#########################################################################
# Methods for deepcopy or factory creation


def init_with_python_lists(cls, *args, **kwargs):
    """
    A factory method to create instances of libfmp4.mpd classes. Should
    always be used instead of the class constructor.

    Usage: to create an mpd data type (i.e. Period) use

    .. code-block:: python

      my_period = create(mpd.Period)

    This method is actually mandatory only for data types that has at
    least one std::vector member, but it is easier to use it everywhere
    than to remember the types for which it is necessary.
    """
    # invoke c++ constructor
    dst = cls(*args, **kwargs)

    # add python members to new object
    copyStdVectors(dst)

    return dst


def deepcopy_align_and_add_python_lists(src):

    # 'align' python list to native c++ members.
    copyPythonToCppMembers(src, remove=False)  # copies lists to std::vectors

    # invoke c++ copy constructor
    dst = type(src)(src)

    # add python members to new object
    copyStdVectors(dst)

    return dst


#########################################################################

#########################################################################
# Methods for class inspection and automatic copy/removal of std:vectors to
# python lists
def getattrFromClassOrObject(type_in, member_name):
    if inspect.isclass(type_in):
        member = getattr(type_in(), member_name)
    else:
        member = getattr(type_in, member_name)

    return member


def _isMemberFromLibfmp4(type_in, member_name):
    member = getattrFromClassOrObject(type_in, member_name)

    # This is the best way I found to accomplish the task. I can't tell if it
    # is very robust.
    try:
        # sometimes __module__ is present and it is None (method-wrapper
        # type)
        attrname = getattr(member, "__module__", "")
        # this instruction is surely not robust, if libfmp4 is imported
        # with "as" under another name, so it is important, if we want to use
        # this approach, that this code is executed in a controlled module
        return attrname and attrname.startswith("libfmp4")
    except Exception as e:
        Context.log_error(
            'Exception {e} while trying to access "__module__" on \
            member {member}'.format(
                e=e, member=member
            )
        )
        return False


def _isMemberAList(type_in, member_name):
    member = getattrFromClassOrObject(type_in, member_name)
    # Even empty lists should be substituted, in case someone wants to append
    # to it.
    # i.e. manifest._utcTimings when empty is just an empty python list
    # and there is no way (that I found) to infer the data type it is
    # supposed to hold. In doubt, you will need to consider this a candidate
    # for "copying".
    return isinstance(member, list)


def _membersShortList(type_in):
    msl = [
        member_name for member_name in dir(type_in) if not member_name.startswith("__")
    ]
    return msl


def _membersToBeProcessed(type_in):
    # The idea is to get the shortest possible list of members that we
    # should recurse into, avoiding the obvious __init__, __ge__ and so on
    # and selecting only lists or types belonging to libfmp4
    return [
        member_name
        for member_name in _membersShortList(type_in)
        if (
            _isMemberAList(type_in, member_name)
            or _isMemberFromLibfmp4(type_in, member_name)
        )
    ]


def copyStdVectors(type_in):
    """
    The idea is to identify every member whose name starts with an underscore
    and add a dynamic attribute with the same name without underscore, i.e.:

    manifest._periods ==> manifest.periods will be added

    We need to handle any possible level of nesting of course, so this method
    will recurse the received type_in.

    The types are a bit peculiar here. We cannot use the standard "inspect"
    method of python because the typical manifest_t data types and related
    do not have all the metadata normally associated with native python types
    such as "__dict__". But we can rely on "__dir__".

    We will use the fact that manifest_t only used stl containers are list
    and will limit recursive iteration to such containers.

    Recursion will stop when all data types of a given member do not belong
    to libfmp4 anymore (i.e. are native C++ types).
    """

    # Iterate over a subset of members that may need to be copied
    for member_name in _membersToBeProcessed(type_in):
        member = getattrFromClassOrObject(type_in, member_name)
        # if the member is a list, recurse on each element
        if isinstance(member, list):
            for item in member:
                Context.log_trace(
                    "copyStdVectors recursing into {} in {}".format(
                        type(item), type(member)
                    )
                )
                copyStdVectors(item)
        else:
            # otherwise just recurse on it
            Context.log_trace("copyStdVectors recursing into {}".format(type(member)))
            copyStdVectors(member)

        # if the member name starts with an underscore, then create a copy
        # i.e.: "instance._member" will have an "instance.member" copy,
        # obtaining at all effects an editable copy of the original data type
        #
        # Notice that we are leveraging the fact that copies of std::vector to
        # python lists (the only real copies we need) are automatically done by 
        # pybind11 anytime you access a std::vector member from python.
        # This is the reason we don't need to use python copy methods here.
        if member_name.startswith("_") and not member_name.startswith("__"):
            # Ok, for the moment we are using this but this is very unsafe.
            # The newly created list is indeed a native python list, but
            # as such it does not have any type hint. User can add
            # whatever type they want and the operation will not fail.
            # But it definitely will screw up things when we try and copy
            # it back in a std::vector which is strictly typed.
            Context.log_trace(
                "copyStdVectors copying {} to {}".format(member_name, member_name[1:])
            )
            setattr(type_in, member_name[1:], member)

    # Additionally, if type_in is AdaptationSet or Representation, copy
    # common members
    if isinstance(type_in, AdaptationSet) or isinstance(type_in, Representation):
        for attr_name in COMMON_MEMBERS:
            setattr(type_in, attr_name, getattr(type_in.common, attr_name))


def _membersToBeRemoved(type_in):
    # All the types that have been dynamically added went into __dict__
    # This is handy.
    return [member_name for member_name in getattr(type_in, "__dict__", {})]


def copyPythonToCppMembers(type_in, remove=True):
    """
    Removes every member present in class' python __dict__.
    Works only if the only containers are list.
    """
    to_be_removed = _membersToBeRemoved(type_in)
    if len(to_be_removed):
        for member_name in to_be_removed:
            member = getattr(type_in, member_name)
            # if the item is a list, recurse and substitute. Only list are
            # supported so far.
            if isinstance(member, list):
                for item in getattr(type_in, member_name):
                    Context.log_trace(
                        "copyPythonToCppMembers recursing into {} "
                        "in {}".format(type(item), type(member))
                    )
                    copyPythonToCppMembers(item)
            else:
                Context.log_trace(
                    "copyPythonToCppMembers recursing into {}".format(type(member))
                )
                copyPythonToCppMembers(member)

            # copy back and remove the attribute
            # This can blow up if containers do not have the same type
            # expected by the std::vector
            Context.log_trace(
                "copyPythonToCppMembers copying {} to {}".format(
                    member_name, "_" + member_name
                )
            )

            # Additionally, if type_in is AdaptationSet or Representation,
            # remove members that had been copied over from common
            if (
                isinstance(type_in, AdaptationSet) or
                isinstance(type_in, Representation)
            ) and member_name in COMMON_MEMBERS:
                dst = getattr(type_in, "common")
                cpp_member_name = member_name
            else:
                dst = type_in
                cpp_member_name = "_" + member_name

            # If everything went right, for every "name" dynamic attribute,
            #  there is a corresponding "_name". If that is not the case, it
            #  means someone has added an additional dynamic attribute,
            #  unrelated to the manipulation we have done in copyPythonMembers.
            if hasattr(dst, cpp_member_name):
                setattr(dst, cpp_member_name, getattr(type_in, member_name))
                Context.log_trace(
                    "copyPythonToCppMembers deleting {}".format(member_name)
                )
                if remove:
                    delattr(type_in, member_name)
            else:
                # This cannot happen because before returing to cpp, all
                #  dynamic members must have been removed. Warn the user.
                Context.log_warning(
                    "copyPythonToCppMembers detected dynamic member {} added "
                    "to manifest. There is no corresponding cpp member! "
                    "Please check that its spelling is correct. "
                    "The member will be removed and ignored.".format(member_name)
                )

                # Best to remove this explicitly, even though I have tested
                #  that pybind11 is kind enough to ignore any dynamic members.
                # Asan tests also show no memory leaks if I avoid deleting
                #  this explicitly.
                if remove:
                    delattr(type_in, member_name)


#########################################################################

# Add "factory" method that should always be used from python to create
#  new instances of libfmp4 types

_thismodule = sys.modules[__name__]

setattr(_thismodule, "create", init_with_python_lists)

#########################################################################


def _get_member_by_tag(cls, tag_name):
    try:
        return getattr(cls, cls.TAGS_MAP[tag_name])
    except KeyError:
        Context.log_error("No mapping for tag {} on type {}".format(tag_name, cls))
    except AttributeError:
        if not hasattr(cls, "TAGS_MAP"):
            Context.log_error("No TAGS_MAP provided for type {}".format(cls))
        else:
            Context.log_error(
                "No attribute {} exists for type {}".format(cls.TAGS_MAP[tag_name], cls)
            )

    return None


def _set_member_by_tag(cls, tag_name, value):
    try:
        return setattr(cls, cls.TAGS_MAP[tag_name], value)
    except KeyError:
        Context.log_error("No mapping for tag {} on type {}".format(tag_name, cls))
    except AttributeError:
        if not hasattr(cls, "TAGS_MAP"):
            Context.log_error("No TAGS_MAP provided for type {}".format(cls))
        else:
            Context.log_error(
                "No attribute {} exists for type {}".format(cls.TAGS_MAP[tag_name], cls)
            )
